#include <stdio.h>
#include <stdlib.h>

/* cf ticket_tab.c */
#include "ticket.h"
#include "ticket_item.h"

/* Structure qui va représenter un maillon de la liste chaînée */
struct ticket_item_list_el_s {
    struct ticket_item_s item; /* les données à stocker */
    struct ticket_item_list_el_s* next_item; /* adresse du maillon suivant */
};

struct ticket_s {
    /* On stocke seulement l'adresse du premier maillon de la liste. On peut
     * maintenant stocker un nombre infini (ou presque...) d'éléments, donc
     * plus besoin d'avoir le nombre d'éléments (qui nous servait seulement à
     * savoir combien on en avait dans le tableau).
     */
    struct ticket_item_list_el_s* first_item;
};

/* cf ticket_tab.c */
static float sum_all_tickets = 0.;

struct ticket_s* ticket_nouveau(void)
{
    struct ticket_s* ticket = malloc(sizeof(struct ticket_s));
    if (ticket == NULL)
    {
        return NULL;
    }

    /* Initialement, le ticket est vide */
    ticket->first_item = NULL;

    return ticket;
}

void ticket_detruire(struct ticket_s* ticket)
{
    /* On a alloué chaque maillon de la chaîne avec un malloc (cf
     * ticket_ajoute_item()), donc il faut libérer chaque maillon, en
     * parcourant la liste.
     */
    struct ticket_item_list_el_s* item = ticket->first_item;

    while (item != NULL)
    {
        struct ticket_item_list_el_s* next_item = item->next_item;
        free(item);
        item = next_item;
    }

    free(ticket);
}

int ticket_ajoute_item(struct ticket_s* ticket, enum item_type_e type, char* description, float prix)
{
    /* Ajoute un élément signifie ajouter un maillon à la liste chaînée. On a
     * besoin d'une adresse pour ce nouveau maillon, adresse qui doit être
     * valide même après l'appel à cette fonction, donc malloc. */
    struct ticket_item_list_el_s* item = malloc(sizeof(struct ticket_item_list_el_s));
    if (item == NULL)
    {
        return 0;
    }

    /* Il faut ajouter ce nouveau maillon comme queue de la liste chaînée, si
     * on souhaite que le ticket s'affiche avec les éléments dans l'ordre
     * d'ajout (oui, ajouter en tête est bien plus simple).
     */

    item->item.type = type;
    item->item.description = description;
    item->item.prix = prix;
    item->next_item = NULL; // ce sera le dernier maillon

    if (ticket->first_item == NULL)
    {
        ticket->first_item = item;
    }
    else
    {
        struct ticket_item_list_el_s* last_item = ticket->first_item;
        while (last_item->next_item != NULL)
        {
            // boucle pour trouver le dernier maillon
            last_item = last_item->next_item;
        }
        last_item->next_item = item;
    }

    return 1;
}

int ticket_supprime_item(struct ticket_s* ticket, int i)
{
    if (ticket->first_item != NULL)
    {
        if (i == 0)
        {
            /* Supprimer le premier élément revient à modifier ticket->first_item */
            struct ticket_item_list_el_s* old_first_item = ticket->first_item;
            ticket->first_item = old_first_item->next_item;

            /* On libère la mémoire de l'élément supprimé */
            free(old_first_item);

            return 1;
        }
        else
        {
            /* Il faut trouver le maillon avant celui à supprimer, pour pouvoir modifier son next_item */
            struct ticket_item_list_el_s* item = ticket->first_item;
            int j = 0;
            while (j < i-1 && item != NULL)
            {
                item = item->next_item;
                j++;
            }

            if (j == i-1 && item != NULL)
            {
                struct ticket_item_list_el_s* item_to_delete = item->next_item;
                item->next_item = item_to_delete->next_item;

                /* On libère la mémoire de l'élément supprimé */
                free(item_to_delete);

                return 1;
            }
        }
    }

    /* Si on arrive ici, c'est qu'on n'a pris de return avant, donc c'est que
     * rien n'a été supprimé, c'est une erreur.
     */
    return 0;
}

void ticket_afficher(struct ticket_s* ticket, FILE* f)
{
    /* cf ticket_tab.c */
    float sum = 0.;
    int nb_items = 0;

    for (struct ticket_item_list_el_s* item = ticket->first_item; item != NULL; item = item->next_item)
    {
        fprintf(f, "%s\t%s\t\t% 2.02f\n", item_type_to_str[item->item.type], item->item.description, item->item.prix);
        sum += item->item.prix;
        nb_items++;
    }

    fprintf(f, "%d item%s\n", nb_items, nb_items > 1 ? "s" : "");
    fprintf(f, "TOTAL: % 2.02f €\n", sum);

    sum_all_tickets += sum;
}

float ticket_somme_tous(void)
{
    return sum_all_tickets;
}
