#include <stdio.h>
#include <stdlib.h>

/* On est dans l'implémentation de l'interface ticket, on doit inclure :
 * - ticket.h parce qu'il contient la définition de enum item_type_e (sans ça,
 *   on ne serait pas obligé de l'inclure) ;
 * - ticket_item.h parce qu'on utilise tout ce qu'il définit.
 *
 * L'ordre d'inclusion des fichiers est important : ticket_item.h a besoin de
 * ticket.h car struct ticket_item_s (dans ticket_item.h) a un membre de type
 * enum item_type_e (défini dans ticket.h).
 */
#include "ticket.h"
#include "ticket_item.h"

/* Dans cette implémentation avec un tableau pour stocker les éléments du
 * tableau, notre tableau va avoir une taille constante. On va utiliser cette
 * constante à plusieurs endroits. Qui dit constante et utilisation à plusieurs
 * endroits, dit #define.
 *
 * Le #ifndef / #endif correspond à l'exercice 3 : on peut définir la valeur de
 * la variable de préprocesseur à la compilation (par exemple :
 * cc -DMAX_NB_TICKET_ITEMS=25 ...
 * ), mais si on ne le fait pas, la valeur par défaut est 10.
 */
#ifndef MAX_NB_TICKET_ITEMS
#define MAX_NB_TICKET_ITEMS 10
#endif

struct ticket_s {
    struct ticket_item_s items[MAX_NB_TICKET_ITEMS];

    /* On a de la place pour stocker au maximum MAX_NB_TICKET_ITEMS, mais on va
     * avoir besoin de savoir combien d'éléments on a déjà sur notre ticket,
     * principalement pour deux raisons :
     * - dans quelle case du tableau on met ce nouvel élément ?
     * - mais au fait, est-ce qu'il nous reste de la place dans le tableau ?
     */
    int nb_items;
};

/* Pour l'exercice 6, on a besoin d'une variable qui va exister le temps de
 * tout le programme, pour stocker au fur et à mesure la somme de tous les
 * tickets. On déclare donc la variable globale sum_all_tickets en-dehors de
 * toute fonction. Le mot clé static permet de limiter la visibilité de la
 * variable à seulement ce fichier.
 */
static float sum_all_tickets = 0.;

struct ticket_s* ticket_nouveau(void)
{
    /* On va renvoyer une adresse vers une zone mémoire qui peut contenir un
     * struct ticket_s. Il faut que cette zone mémoire reste valide une fois
     * que cette fonction a retourné, on va donc faire un malloc.
     */
    struct ticket_s* ticket = malloc(sizeof(struct ticket_s));
    /* malloc peut échouer, il est donc importer de tester si on a vraiment
     * récupéré une zone mémoire *avant* de tenter d'accéder à la mémoire, en
     * modifiant un champ de la structure par exemple.
     */
    if (ticket == NULL)
    {
        return NULL;
    }

    ticket->nb_items = 0;

    return ticket;
}

void ticket_detruire(struct ticket_s* ticket)
{
    /* 1 malloc = 1 free */
    free(ticket);
}

int ticket_ajoute_item(struct ticket_s* ticket, enum item_type_e type, char* description, float prix)
{
    /* On vérifie que le tableau ne soit pas déjà plein. */
    if (ticket->nb_items == MAX_NB_TICKET_ITEMS)
    {
        return 0;
    }

    /* Ici, on copie l'adresse de la chaîne de caractères stockée dans
     * description, pas la chaîne de caractères elle-même. Il faut donc que
     * cette adresse soit valide le temps de l'utilisation du ticket, ce qui
     * est le cas avec des chaînes de caractères constantes.
     */
    ticket->items[ticket->nb_items].description = description;
    ticket->items[ticket->nb_items].type = type;
    ticket->items[ticket->nb_items].prix = prix;
    ticket->nb_items++;

    return 1;
}

/* Exercice 5 */
int ticket_supprime_item(struct ticket_s* ticket, int i)
{
    /* On vérifie si on a vraiment un élément qui correspond à cet indice */
    if (i < 0 || i >= ticket->nb_items)
    {
        return 0;
    }

    /* La suppression consiste à enlever l'élément du tableau, mais on ne peut
     * pas avoir de "trou" dans le tableau (comment savoir que c'est un trou ?).
     * Donc on comble le trou en décalant tous les éléments qui suivent celui
     * qu'on souhaite supprimer (l'élémént i+1 prend la place de l'élément i,
     * l'élémént i+2 prend la place de l'élément i+1, etc).
     */
    for (int j = i+1; j < ticket->nb_items; j++)
    {
        ticket->items[j-1] = ticket->items[j];
    }

    ticket->nb_items--;

    return 1;
}

void ticket_afficher(struct ticket_s* ticket, FILE* f)
{
    float sum = 0.;
    for (int i = 0; i < ticket->nb_items; i++)
    {
        /* 2.02f pour formatter correctement l'affichage des prix, avec
         * toujours 2 décimales.
         * item_type_to_str (dans ticket_item.h) nous permet de facilement
         * convertir la valeur de type enum_item_type_e en chaîne de
         * caractères.
         * '\t' correspond au caractère tabulation, pour faire des espaces de
         * façon à peu près standard.
         *
         * Exercice 4 : utiliser fprint(f, ...) au lieu de printf(...) permet
         * d'écrire sur le descripteur de fichier que l'utilisateur nous a
         * passé en paramètre : ça peut être stdout (la sortie standard = le
         * terminal) ou bien un fichier.
         */
        fprintf(
            f,
            "%s\t%s\t\t% 2.02f\n",
            item_type_to_str[ticket->items[i].type],
            ticket->items[i].description,
            ticket->items[i].prix
        );

        /* On profite de la boucle sur tous les éléments pour calculer la somme
         * du ticket (pas la peine de faire une seconde boucle par la suite !).
         */
        sum += ticket->items[i].prix;
    }

    /* Condition ternaire pour accorder correctement en nombre le mot "item" */
    fprintf(f, "%d item%s\n", ticket->nb_items, ticket->nb_items > 1 ? "s" : "");
    fprintf(f, "TOTAL: % 2.02f €\n", sum);

    /* Exercice 6 : on a affiché le ticket, ajoutons son montant à la somme de tous les tickets. */
    sum_all_tickets += sum;
}

/* Exercice 6 */
float ticket_somme_tous(void)
{
    return sum_all_tickets;
}
